

class PDFLexerException(Exception):
    pass


class PDFLexerEOLException(Exception):
    pass


class PDFObjectTypeException(Exception):
    pass


class PDFParserException(Exception):
    pass


class PDFParserDoneException(Exception):
    pass


class PDFObjException(Exception):
    pass


class PDFObjNotFoundException(Exception):
    pass


class PDFStreamException(Exception):
    pass


class PDFDocumentException(Exception):
    pass


class PDFSyntaxException(Exception):
    pass


class PDFInterpreterException(Exception):
    pass


class PDFCMapNotFoundException(Exception):
    pass
